import numpy as np
import pandas as pd

class MyLinearRegression:

    def __init__(self, regularization=None, n_iterations=1_000_000, learning_rate=1e-4, reg_alpha=1e-5, verbose=False): # __init__ is the constructor function in python.
        self.regularization=regularization
        self.W = None
        self.n_iterations = n_iterations
        self.learning_rate = learning_rate
        self.reg_alpha = reg_alpha
        self.verbose = verbose
    
    def _prepare_X(self, X):
        # Transform into numpy array if it is not one yet and adds the bias
        if type(X) == pd.DataFrame or type(X) == pd.Series:
            X = X.to_numpy()
        
        bias = np.ones((X.shape[0], 1)) # value of 1 for all instances
        
        if np.all(X[:,-1] - bias == 0): # If the last column is already the bias, we don't change it
            return X

        X = np.append(X, bias, 1) # add bias as an extra column to X
        
        return X
        
    def _prepare_y(self, y):
        # Transforms to numpy array if it is not already and reshapes to have 2 dimensions
        if type(y) == pd.DataFrame or type(y) == pd.Series:
            y = y.to_numpy()

        if len(y.shape) == 1:
            y = y.reshape(-1, 1)
            
        return y
    
    def compute_cost(self, X, y, y_pred):
        # mse: 1/2m * SUM((y-y_hat)^2)
        mse = #complete the formula
        
        return mse + self.get_regularization_term()
    
    def get_derivative(self, X, y, y_pred):
        # derivative will be 1/m * SUM((y-y_pred)*X)
        #implement this function
        return 
    
    def get_regularization_term(self):
        reg = 0
        if self.regularization == 'l2':
            reg = # complete this part of the code with formula of computing the l2 regularization term
        
        if self.regularization == 'l1':
            reg = # complete this part of the code with formula of computing the l1 regularization term
        
        return self.reg_alpha * reg
    
    def get_derivative_for_regularization_term(self):
        reg = 0
        if self.regularization == 'l2':
            reg = 2*self.W.reshape(-1, 1)
        
        if self.regularization == 'l1':
            reg = 1
        
        return self.reg_alpha * reg
    
    def get_gradient_update(self, X, y, y_pred):
        derivative = self.get_derivative(X, y, y_pred)
        
        return derivative.reshape(-1, 1) + self.get_derivative_for_regularization_term()
        
        
    def fit(self, X, y):
        # We add the bias to X and set the shape of y to be what we need for fitting.
        X = self._prepare_X(X)
        y = self._prepare_y(y)
        
        # We initialize the weights to be a normal distribution centered on 0
        self.W = np.random.normal(0, 0.2, (X.shape[1], y.shape[1]))
        
        previous_cost = None # we keep track of the previous cost so we stop if the cost is not improving
        
        for it in range(self.n_iterations):
            y_pred = self.predict(X)
            cost = self.compute_cost(X, y, y_pred)
            if previous_cost is not None and abs(cost - previous_cost) < 1e-8: # Dynamically adjust the learning rate
                self.learning_rate /= 2
                if self.learning_rate < 1e-10:
                    print(f"Stopped at iteration {it} with LR:{self.learning_rate} and cost:{cost}")
                    break
            
            previous_cost = cost # update previous cost to store the new cost
            
            if it % 1000 == 0 and self.verbose: # For us to keep track of how it is evolving over time.
                print(f"Iteration: {it} Cost: {cost}")
            gradient_update = self.learning_rate * self.get_gradient_update(X, y, y_pred).reshape(-1, 1)
    
            self.W = self.W - gradient_update
            
        print(f"MSE:{previous_cost}") 

    def predict(self, x):
        return # complete this part of the code
